local shared = ...

local canvas_update_queue = {}

local function add_canvas_to_update_queue(canvas)
    for _, item in ipairs(canvas_update_queue) do
        if item == canvas then
            return
        end
    end
    table.insert(canvas_update_queue, canvas)
end

local CanvasEntity = {
    initial_properties = {
        visual = "upright_sprite",
        pointable = false,
        physical = false,
        static_save = false,
    },
}

local function validate_vec2(data)
    return type(data) == "table" and
        type(data.x) == "number" and
        type(data.y) == "number"
end
local function validate_vec3(data)
    return validate_vec2(data) and
        type(data.z) == "number"
end

local function validate_staticdata(data)
    return
        type(data) == "table" and
        -- allow canvases created with older versions of the mod
        (data.node_offset == nil or validate_vec3(data.node_offset)) and
        validate_vec2(data.size) and
        validate_vec2(data.bitmap_size) and
        type(data.bitmap) == "table" and
        #data.bitmap == data.bitmap_size.x * data.bitmap_size.y
end

function CanvasEntity:on_activate(staticdata)
    self.object:set_armor_groups({ immortal = 1 })
end

function CanvasEntity:setup(node_pos, size, bitmap_size)
    self.node_offset = node_pos - self.object:get_pos()
    self.size = size
    self.bitmap_size = bitmap_size

    self.bitmap = {}
    for i = 1, self.bitmap_size.x * self.bitmap_size.y do
        self.bitmap[i] = shared.TRANSPARENT
    end
    
    self:add_canvas_at_pos(node_pos)
end

function CanvasEntity:update_later()
    add_canvas_to_update_queue(self)
end

function CanvasEntity:update_immediately()
    local png = core.encode_base64(
        core.encode_png(self.bitmap_size.x, self.bitmap_size.y, self.bitmap, 9)
    )
    self.object:set_properties({
        visual_size = vector.new(self.size.x, self.size.y, 0),
        textures = {
            "[png:" .. png,
            -- "blood_splatter_debug_coordinates.png^([png:" .. png .. "^[resize:128x128)",
            "blank.png",
        },
    })
    self:reset_remove_timer()
end

local function clamp(val, min, max)
    return math.min(math.max(val, min), max)
end

-- "x" and "y" parameters are always 0-based in the functions below.

-- "out-of-bounds on bitmap" doesn't always also mean "out-of-bounds in list",
-- so there's a separate way to check.
function CanvasEntity:pos_in_bounds(x, y)
    return x >= 0 and x <= self.bitmap_size.x - 1 and
        y >= 0 and y <= self.bitmap_size.y - 1
end

function CanvasEntity:pos_get_index(x, y)
    return y * self.bitmap_size.x + x + 1
end

-- If the pixel is outside the canvas area, nothing will be drawn.
function CanvasEntity:draw_pixel(x, y, color, remover)
    if not self:pos_in_bounds(x, y) then
        return
    end
    local index = self:pos_get_index(x, y)

    if self.bitmap[index] ~= color then
        self.bitmap[index] = color

        if remover and self:is_empty() then
            self.object:remove()
        else
            self:update_later()
        end
    end
end

-- To get correct results, at least one pixel of the rectangle must be inside
-- the canvas area. If this requirement is not fulfilled, the result will be
-- incorrect (but there will be no errors / data invalidation).
function CanvasEntity:draw_rect(x, y, size, color, remover)
    assert(size >= 1)

    local max_x = self.bitmap_size.x - 1
    local max_y = self.bitmap_size.y - 1
    local x1, y1 =
        clamp(x, 0, max_x),
        clamp(y, 0, max_y)
    local x2, y2 =
        clamp(x + size - 1, 0, max_x),
        clamp(y + size - 1, 0, max_y)

    -- TODO: make this more fancy using stride stuff (get rid of pos_get_index)
    for yy = y1, y2 do
        for xx = x1, x2 do
            self.bitmap[self:pos_get_index(xx, yy)] = color
        end
    end

    if remover and self:is_empty() then
        self.object:remove()
    else
        self:update_later()
    end
end

function CanvasEntity:is_empty()
    for _, c in ipairs(self.bitmap) do
        if c ~= shared.TRANSPARENT then
            return false
        end
    end
    return true
end

function CanvasEntity:get_staticdata()
    return core.serialize({
        node_offset = self.node_offset,
        size = self.size,
        bitmap_size = self.bitmap_size,
        bitmap = self.bitmap,
    })
end

function CanvasEntity:get_node_pos()
    return (self.object:get_pos() + self.node_offset):round()
end

core.register_entity("blood_splatter:canvas", CanvasEntity)

function shared.update_canvases()
    for _, canvas in ipairs(canvas_update_queue) do
        -- This doesn't cause a crash if the canvas entity has been removed.
        canvas:update_immediately()
    end
    canvas_update_queue = {}
end


--code below CC0 by shaft
local decalLifetime = tonumber(core.settings:get("blood_splatter.decalLifetime")) or 300
local canvas_positions = {}
local function job_remove_canvas(self)
	self.object:remove()
end

function CanvasEntity:add_canvas_at_pos(node_pos)
	self.node_pos = node_pos
	self.time = os.clock()
	self.lifetime = core.after(decalLifetime, job_remove_canvas, self)
	local bitpacked_pos = core.hash_node_position(self.node_pos)
	if not canvas_positions[bitpacked_pos] then canvas_positions[bitpacked_pos] = {} end
	table.insert(canvas_positions[bitpacked_pos], self)
end

function CanvasEntity:remove_canvas_from_pos()
	local bitpacked_pos = core.hash_node_position(self.node_pos)
	local entries = canvas_positions[bitpacked_pos]
	for i,v in ipairs(entries) do
		if v == self then
			v.lifetime:cancel()
			table.remove(entries,i)
			break
		end
	end
	
	if #entries == 0 then canvas_positions[bitpacked_pos] = nil end
end

function CanvasEntity:on_deactivate()
	self:remove_canvas_from_pos()
end

function CanvasEntity:reset_remove_timer()
	local time = os.clock()
	if self.time + 1 < time then
		local bitpacked_pos = core.hash_node_position(self.node_pos)
		local entries = canvas_positions[bitpacked_pos]
		for _,v in ipairs(entries) do
			if v == self then
				v.lifetime:cancel()
				v.lifetime = core.after(decalLifetime, job_remove_canvas, self)
				break
			end
		end
	end
	self.time = time
end

core.register_on_dignode(function(pos)
	local bitpacked_pos = core.hash_node_position(pos)
	local entries = canvas_positions[bitpacked_pos]
	if entries then
		for i=#entries,1,-1 do
			entries[i].object:remove()
		end
	end
end)
--end of code by shaft